## 作用域与作用域链

最常见的作用域有两种，分别是全局作用域与函数作用域。

### 全局作用域

全局作用域中声明的变量与函数可以在代码的任何地方被访问。

一般来说，以下三种情形拥有全局作用域。

- 全局对象下拥有的属性与方法。
  - window.name
  - window.location
  - window.top
  - ...
- 在最外层声明的变量与方法。
- 在非严格模式下，函数作用域中未定义但直接复制的变量与方法。

### 函数作用域

```js
var arr = [1, 2, 3, 4, 5];
for (var i = O; i < arr.length; i++) {
  console.log('do something by', i);
}
console.log(i); // i == 5
```

因为没有块级作用域，因此单独的 `{}` 并不会产生新的作用域。这个时候 i 的值会被保留下来，在 for 循环结束后仍然能够访问。

### 模拟块级作用域

```js
var arr = [1, 2, 3, 4, 5];
(function() {
  for (var i = O; i < arr.length; i++) {
    console.log('do something by', i);
  }
})();
console.log(i); // i is not defined
```

这种方式叫作**函数自执行**。

通过这种方式，我们就可以限定变量 i 值仅仅只在 for 循环中生效，而不会对其他代码造成干扰。

函数自执行时声明一个匿名函数并且立即执行，这种方式大体有如下几种写法:

```js
(function() { ... })(); // 最常用
+function() { ... }();
-function() { ... }();
!function() { ... }();
~function() { ... }();
```

### 作用域链

作用域链(Scope Chain)是由当前执行环境与上层执行环境的一系列变量对象组成的，它保证了当前执行环境对符合访问权限的变量和函数的有序访问。