运行环境

·知识点：
1.页面加载过程
2.性能优化
3.安全性

####页面加载过程
·题目：
  ·从输入url到得到html的详细过程
  ·window.onload和DOMContentLoaded的区别
·知识点：
  ·加载资源的形式
    ·输入url（或跳转页面）加载html
    ·http://coding.com
    ·加载html中的静态资源
    ·<script src="/static/js/jquery.js"></script>
  ·加载一个资源的过程
    ·浏览器根据DNS服务器得到域名的IP地址
    ·向这个IP的机器发送http请求
    ·服务器收到、并处理返回的http请求
    ·浏览器得到返回内容
  ·浏览器渲染页面的过程
    ·根据html结构生成DOM Tree（只是一个树形结构）
    ·根据css生成CSSOM（css静态化代码进行结构化处理）
    ·将DOM和CSSOM整合形成RenderTree（渲染树）
    ·根据RenderTree开始渲染和展示
    ·遇到script时，会执行并阻塞渲染（js有权利改变dom结构）
思考：
·为什么把css放到head中
·为什么把js放到body最下面
  ·防止阻塞，让页面更快展示出来
  ·有可能拿不到dom节点
  
·解答：
·window.onload和DOMContentLoaded的区别（尽量使用第二种）
  window.addEventListener('load',function(){
    //页面的全部资源加载完才会执行，包括图片、视频等
  })
  document.addEventListener('DOMContentLoaded',function(){
    //DOM渲染完即可执行，此时图片、视频可能还没有加载完
  })

####性能优化
·原则：
  ·多使用内存、缓存或者其他方法
  ·减少CPU计算、减少网络请求、减少IO
·从哪里入手
  ·加载页面和静态资源
    ·加载资源优化
      ·静态资源的压缩合并
      ·静态资源缓存
      ·使用CDN让资源加载更快
      ·使用SSR后端渲染，数据直接输出到HTML中
  ·页面渲染
    ·渲染优化
      ·CSS放前面，JS放后面
      ·懒加载（图片懒加载、下拉加载更多）
      ·减少DOM查询，对DOM查询做缓存
      ·减少DOM操作，多个操作尽量合并在一起执行
      ·事件节流
      ·尽早执行操作（如DOMContentLoaded）
  ·资源合并
  ·缓存
    ·通过连接名称控制缓存
    ·<script src="abc_1.js"></script>
    ·只有内容改变的时候，链接名称才会改变
    ·<script src="abc_2.js"></script>
  ·CDN
  ·使用SSR后端渲染
    ·现在vue react提出了这样的概念
    ·其实jsp php asp都属于后端渲染
  ·懒加载
    <img id="img1" src="img.png" data-realsrc="abc.png" alt="">
    <script type="text/javascript">
      var img1 = document.getElementById('img1')
      img1.src = img1.getAttribute('data-realsrc')
    </script>
  ·缓存DOM查询
    // 未缓存DOM查询
    var i 
    for(i=0;i<document.getElementByTagName('p').length;i++){
      //todo
    }
    // 缓存了DOM查询
    var pList = document.getElementByTagName('p')
    var i
    for(i=0;i<pList.length;i++){
      //todo
    }
  ·合并DOM插入
    var listNode = document.getElementById('list')
    // 要插入10个li标签
    var frag = document.createDocumentFragment();
    var x,li;
    for(x=0;x<10;x++){
      li = document.createElement('li');
      li.innerHTML = 'List item'+x;
      frag.appendChild(li);
    }
    listNode.appendChild(frag);
  ·事件节流
    var textarea = document.getElementById('text')
    var timeoutId
    textarea.addEventListener('keyup',function(){
      if(timeoutId){
        clearTimeout(timeoutId)
      }
      timeoutId = setTimeout(function(){
        //触发change事件
      },100)
    })
  ·尽早操作
    window.addEventListener('load',function(){
      //页面的全部资源加载完才会执行，包括图片、视频等
    })
    document.addEventListener('DOMContentLoaded',function(){
      //dom渲染完即可执行，此时图片，视频还可能没有加载完
    })

####安全性
1.XSS跨站请求攻击
2.XSRF跨站请求伪造

·XSS
  ·在新浪博客写一篇文章，同时偷偷插入一段script
  ·攻击代码中，获取cookie，发送 自己的服务器
  ·发布博客，有人查看博客内容
  ·会把查看者的cookie发送到攻击者的服务器
解决：·前端替换关键字<>为后端替换

·XSRF
  ·你已登陆一个购物网站，正在浏览商品
  ·该网站付费接口是xxx.com/pay?id=1但是没有任何验证
  ·然后你收到一封邮件，隐藏着<img src="xxx.com/pay?id=1">
  ·你查看邮件的时候，就已经悄悄的付费购买了
解决：·增加验证流程，如输入指纹，密码，短信验证码