# 大纲

- 早期的双飞翼布局 + CSS HACK
- 基于移动端的px与REM转换兼容方案
- 弹性盒模型与Reset选择
- 自制的ICON-FONT与常用字体排版
- CSS代码检测与团队项目规范
- CSS绘制特殊图形 高级技巧
- BFC IFC GFC FFC
- CSS Grid-Layout
- CSS分层与面向对象

## 双飞翼布局

- position
- float
- 父边距
- 等高
- 盒子模型
- 清除浮动

### IE6经典BUG

#### 1.IE6怪异解析之padding与border算入宽高

> 解决方案：加入文档声明`<!doctype html>`


#### 2.IE6在块元素、左右浮动、设定margin时造成margin双倍(双倍边距)

> 解决方案：`display：inline`

#### 3.以下三种其实是同一种bug，其实也不算是个bug，举个例子：父标签高度20，子标签11，垂直居中，20-11=9，9要分给文字的上面与下面，怎么分？IE6就会与其它的不同，所以，尽量避免。

1. 字体大小为奇数之边框高度少1px
   > 解决方法：字体大小设置为偶数或`line-height`为偶数
2. line-height，文本垂直居中差1px
   > 解决方法：`padding-top`代替`line-height`居中，或`line-height`加1或减1
3. 与父标签的宽度的奇偶不同的居中造成1px的偏离
   > 解决方法：如果父标签是奇数宽度，则子标签也用奇数宽度;如果是父标签偶数宽度，则子标签也用偶数宽度。

#### 4.内部盒模型超出父级时，父级被撑大。

> 解决方法：父标签使用`overflow:hidden`

#### 5.line-height默认行高bug

> 解决方法：`line-height`设值

#### 6.行标签之间会有一小段空白

> 解决方法：`float`或结构并排(可读性差，不建议)

#### 7.标签高度无法小于19px

> 解决方法：`overflow: hidden;`

#### 8.左浮元素margin-bottom失效

> 解决方法：显示设置高度 or 父标签设置`_padding-bottom`代替子标签的`margin-bottom` or 再放个标签让父标签浮动，子标签 `margin-bottom`，即(`margin-bottom`与`float`不同时作用于一个标签)

#### 9.img于块元素中，底边多出空白

> 解决方法：父级设置`overflow: hidden;` 或 `img { display: block; }` 或 `_margin: -5px;`

#### 10.li之间会有间距

> 解决方法：`float: left;`

#### 11.块元素中有文字及右浮动的行元素，行元素换行

> 解决方法：将行元素置于块元素内的文字前

#### 12.position下的left，bottom错位

> 解决方法：为父级(relative层)设置宽高或添加`*zoom:1`

#### 13.子级中有设置position，则父级overflow失效

> 解决方法：为父级设置`position:relative`

## 弹性盒模型与Reset的选择

- flex 模型
- *的杀伤力太大
- Reset.css 重置 Normalize.css 修复 Neat.css 融合
- `html {box-sizing: border-box;} *,*:before,*:after {box-sizing: border-box}`(移动端布局一定要加)

## ICON-FONT 与常用字体排版

- no-image时代 不超过纯色为2的图像
- 宋体非宋体 黑体非黑体 Windows下的宋体叫中易黑体SimSun，Mac是华文宋体STSong。Windows下的黑体叫中易黑体SimHei, Mac是华文黑体STHeti。
- 不要只写中文字体名，保证西文字体在中文字体前面。 Mac -> Linux -> Windows
- 切忌不要直接使用设计师 PSD 的设计 font-family, 关键时刻再去启动 `font-face(typo.css、Entry.css、Type.css)`
- `font-family: sans-serif`
- [cssicon](http://cssicon.space/#/)

## 有关设计的JS

- underline.js
- responsify.js
- typedetail.com
- cssicon.space
- designreseach.space

## CSS代码检测与团队项目规范(HINT.css)

- 不要使用多个class选择元素，如a.foo.boo，这在ie6及以下不能正确解析
- 移除空的css规则，如a{}
- 正确的使用显示属性，如display:inline不要和width，height，float，margin,padding同时使用，display:inline-block不要和float同时使用等
- 避免过多的浮动，当浮动次数超过十次时，会显示警告
- 避免使用过多的字号，当字号声明超过十种时，显示警告
- 避免使用过多web字体，当使用超过五次时，显示警告
- 避免使用id作为样式选择器
- 标题元素只定义一次
- 使用width:100%时要小心
- 属性值为0时不要写单位
p 各浏览器专属的css属性要有规范，
  例如.foo{-moz-border-radius:5px;border-radius:5px}
- 避免使用看起来像正则表达式的css3选择器
- 遵守盒模型规则

## 文档布局对象

Box是CSS布局对象的基本单位，直观的说，就是一个页面是由很多个Box组成的，元素的类型和display属性，决定了这个Box的类型。不同类型的Box，会参与到不同的Formatting Context(一个决定如何渲染文档的容器)，因此Box内的元素会以不同的方式渲染。

#### 两种box模型

1. **block-level box** ： display属性为block、list-item、table、inline-block的元素。一旦一个元素形成了块级，就会生成一个 block-level box，并参与到block fomatting context中去。
2. **inline-level box**：display属性为inline、inline-table的元素，会生成inline-level box 并且参与到inline formatting Context中。

#### 什么是Formatting Context

> Formatting Context 是W3C CSS2.规范中的一个概念。它是页面中一块渲染区域，并且有一套渲染规则，它决定了其子元素将如何定位以及和其他元素的关系和相互作用。最常见的Formatting Context 有Block Formatting Context（简称BFC）和Inline Formatting Context（简称IFC）

**会生成BFC的元素**

- 根元素
- float属性不为none
- position为absolute或fixed
- display为inline-block, table-cell, table-caption, flex, inline-flex
- overflow不为visible

#### 总结

- BFC就是页面上的一个隔离的独立容器，容器里面的子元素不会影响到外面的元素。
- IFC（Inline Formatting Contexts）行级格式化上下文，IFC的line box（线框）高度由其包含行内元素中最高的实际高度计算而来（不受到竖直党项的padding和margin影响）
- FFC（Flex Formatting Contexts） 自适应格式化上下文，display值为flex或者inline-flex的元素会生成自适应容器
- GFC（GridLayout Formatting Contexts）网格布局格式化上下文，当为一个元素设置display值为frid的时候，此元素将会获得一个独立的渲染区域我们可以通过网格容器(frid container)上定义网格、定义行（grid definition rows） 和网格定义列(grid definition columns)属性各在网格项目（grid item）上定义网格行（gird row）和网格列 （grid columns） 为每一个网格项目（dird item）定义位置和空间。

## CSS分层与面向对象

- SMACSS
- BEM
- SUIT
- ACSS
- ITCSS
