## 你不知道的clip-path

在我们想要做网页上的一些特殊形状的时候，我们都知道可以用border-radius、border、box-shadow、outline、伪元素等等来去组合成各种形状，以达到自己想要的结果。这明显的弊端是增加了我们代码量的同时又增加了元素的数量。于是我们在致力于寻找用一行CSS代码就可以画出各种各样的图形的方案。终于，clip-path出现了。

### clip-path

> clip-path前身是[clip](https://www.w3cplus.com/css3/clip.html)。

**裁切路径和裁切区域**

在clip-path上有两个非常重要的概念：裁切路径（Clipping Path）和裁切区域（Clipping Region）。这也正如clip-path单词所示，它的作用主要是依据所指定的路径进行裁切，而裁切区域就是我们要保留的区域。

**clip-path的语法**

```css
clip-path: <clip-source> | [ <basic-shape> || <geometry-box> ] | none
```

默认值为none

1. clip-source：可以是内、外部的SVG的元素的URL引用
2. basic-shape: 使用一些基本的形状函数创建的一个形状。主要包括circle()、ellipse()、inset()和polygon()。
3. geometry-box: 是可选参数。此参数和basic-shape函数一起使用时，可以为basic-shape的裁剪工作提供参考盒子。

**clip-path的使用**

```css
img {
    clip-path: polygon(50% 0%, 100% 50%, 50% 100%, 0% 50%);
}
```

polygon函数作用是绘制多边形，以点连接成路径进行剪切（路径为点之间的顺时针连接）。

**利用geometry-box来决定剪切方式**

geometry-box可以是shape-box、fill、stroke或者view-box。其中shape-box应用于HTML元素，它具有四种值：margin-box、border-box、padding-box和content-box。

```css
.clip-me {
    clip-path: polygon(10% 20%, 20% 30%, 50% 80%) margin-box;
    margin: 10%;
}
```

## transition动画

tansition又称补间动画。指的是定义动画结尾的状态，此属性可以补充中间发生的动作从而形成一个动画。

#### transition的值

- transition-property：指CSS属性的name，赋予transition效果。
- transition-duration：指transition动画需要多少秒或者多少毫秒完成。
- transition-timing-function：指定transition效果的转速曲线。
- transition-delay：指定transition效果开始的时候，也就是延迟效果。
- 混合写法。

```css
div{
    width:100px;
    height:100px;
    background-color:red;
    transition:width 1s ease-in;
}
div:hover{
    width:300px;
}
```

## animation动画

animation是自定义动画，通过@Keyframes定义动画的运动状态。

#### animation的值
1. animation-name：规定需要绑定到选择器的keyframe名称。
2. animation-duration：规定完成动画所花费的时间，以秒或毫秒计。
3. animation-timing-function：规定动画的速度曲线。
4. animation-delay：规定动画开始的延迟。
5. animation-iteration-count：规定动画应该播放的次数。
6. animation-direction：规定是否轮流反向播放动画。

#### @Keyframes规则

@Keyframes作用是animation自定义动画的运动状态，也就是说，它的作用是创建动画。

**动画的原理**：原理就是将一套CSS样式之间变化为另一套样式。

以百分比来规定改变发生的时间，或者通过关键词‘from’和‘to’，等价于0%和100%。0%是动画的开始时间，100%是动画的结束时间。

```css
.test{
    width:200px;
    height:200px;
    background:yellowgreen;
    border-radius:100px;
    animation:bounce 3s;
}
@keyframes bounce{
    60% ,80% ,to {
        transform:translateY(400px);
    }
    70%{
        transform: translateY(300px);
    }
    90%{
        transform:translateY(300px);
    }
}
```

## 贝塞尔曲线

上面说了`timing-function`指定动画的速度曲线，到底这个是什么，将在这里揭晓。

`animation-timing-function`和`transition-timing-function`两个属性提供动画的速度曲线，分别提供了`ease、liner、ease-out、ease-in-out`这些预设速度。同时还可以通过贝塞尔曲线来自己定义。

**贝塞尔曲线的原理：** 斜率不变，则速度恒定。通过改变运动状态的速度，从而形成各种运动的效果。

**贝塞尔曲线的规则：** `cubic-bezier (x1,y1,x2,y2)`

## transform

transform可以定义盒子的形态变化。主要有一下变化

1. translate：定义盒子的位置变化（移动）。

```css
div{
    width:100px;
    height:100px;
    background-color:red;
    /*向X轴正向移动200px*/
    transform:translateX(200px);
}
```

2. scale：定义盒子的缩放变化。

```css
div{
    width:100px;
    height:100px;
    background-color:red;
    /*缩放2倍*/
    transform:scale(2);
}
```

3. rotate：定义盒子的旋转变化。

```css
div{
    width:100px;
    height:100px;
    background-color:red;
    /*向Z轴方向旋转45度*/
    transform:rotateZ(45deg);
}
```

4. skew：定义盒子的倾斜变化。

```css
div{
    width:100px;
    height:100px;
    background-color:red;
    /*向X轴方向倾斜45度*/
    transform:skew(45deg);
}
```

### transform-style

transform-style定义了变换是什么类型的变换，当值为preserve-3d的时候，图形的变换就变成了3D变换

```css
div{
    width:100px;
    height:100px;
    background-color:red;
    transform-style:preserve-3d;
}
```

### prespective

prespective是景深的意思，同样也叫做视距，表示用户视觉到物理的距离，达到视觉上3D的效果。

可以通过perspective-origin改变视距的方向

```css
div{
    /*值的书写*/
    /*
    x-position
        指定消失点的横坐标，其值有以下形式：
        <percentage> 百分比，相对于元素宽度，可为负值。
        <length> 长度值，可为负值。
        left，关键字，0值的简记。
        center，关键字，50%的简记。
        right，关键字，100%的简记。
    y-position
        指定消失点的纵坐标，其值有以下形式：
        <percentage> 百分比，相对于元素的高度，可为负值。
        <length> 长度值，可为负值。
        top，关键字，0值得简记。
        center，关键字，50%的简记。
        bottom，关键字，100%的简记。

    */
}
```

## Matrix

> 在数学中，矩阵（Matrix）是一个按照长方阵列排列的复数或实数集合，最早来自于方程组的系数及常数所构成的方阵。矩阵是高等代数学中的常见工具，也常见于统计分析等应用数学学科中。在物理学中，矩阵于电路学、力学、光学和量子物理中都有应用；计算机科学中，三维动画制作也需要用到矩阵。矩阵的运算是数值分析领域的重要问题。由 m × n 个数aij排成的m行n列的数表称为m行n列的矩阵，简称m ×n矩阵。这m×n个数称为矩阵A的元素，简称为元。
