```js
var a = 25;
(function() {
  alert(a);
  var a = 30;
})()

// undefined
```

**为什么a是undefined?**

- 立即函数
- 函数首先是匿名的
- 由于 JS 内声明变量会向前提升,上述函数内的代码等价于

```js
var a;
alert(a);	// undefined
a = 30;
```

写一个立即执行函数保证自己的内部代码不被外部所污染

```js
this.a = 20;

var test = {
  a: 40,
  init: () => {
    console.log(this.a);
    function go() {
      this.a = 60;
      console.log(this.a);
    }
    go.prototype.a = 50;
    return go;
  }
};

var p = test.init();
p();
new(test.init())();
```

**什么是闭包？**

**闭包为什么会造成内存泄露**

```js
function f1() {
  var N = 0;
  function f2() {
    N += 1;
    console.log(N)
  }
  return f2;	// 在f2的内部得到了N
}
var result = f1();
result();
result();
result();
result = null; // 使用完以后对result赋值为null， 这样就不会造成内存泄露了
// 不仅仅能拿到变量，变量还一直在内存中
```

```js
function Product() {
  var name;
  this.setName = function(value) {
    name = value;
  };
  this.getName = function() {
    return name;
  }
}

var s = new Product();
s.setName('hello');
```

**原型链**

**面向对象编程是什么? 怎么实现面向对象编程？**

```js
var Car = function(color) {
  // Car.prototype.constructor = Car
  // 构造函数和初始化这个类就是一个东西了
  this.color = color;
  this.sail = function() {
    console.log(this.color + "卖13W");
  }
};

Car.prototype.sail = function() {
  console.log(this.color + "卖13W");
};

// Todo 构造函数中的方法 和 构造函数原型上方法的区别

var BMW = function(color) {
  Car.call(this.color)
};

// 面试 必问
// 子类
// 1. 拿到父类原型类上的方法
// 2. 不能让构造函数执行两次
// 3. 引用的原型链不能是按值引用
// 4. 修正子类的 constructor
var __pro = Object.create(Car.prototype);
__pro.constructor = BMW;
BMW.prototype = __pro;


// BMW.prototype = Car.prototype;

// new 的话是把原型链上的东西都拿过来，
// call 的话是改变 this 的指向
var s = new Car();

console.log(s);
```

**this 要用熟**

```js
// 函数提升
(function() {
  var a = 20;
  function a() {};
  console.log(a);

  // 函数提升的优先级更高。所以这段代码是这样的
  function a() {}
  var a;
  a = 20;
  console.log(a);	// 20
})();
```

```js
(function() {
  var a = 20;
  var b = c = a;
  console.log(c);
})();

alert(c);

// 上述代码相当于
(function() {
  var a = 20;
  var b = a;
  c = a;
})();

console.log(c);	// 20
```

```js
function test() {
  this.a = 20;
}
this.prototype.a = 30;
var q = new test;
alert(q.a)
// 构造函数内的优先级要比原型链内的优先级要高。
```

```js
var user = {
  age: 20,
  init: function() {
    console.log(this.age);
  }
};

var data = { age: 40 };
// bind 以后返回的是一个新函数
var s = user.init.bind(data);
s(); // 40
user.init();	// 20
```

## 知识点
es5 中能问到的就是这些东西

1. 立即执行函数
2. 闭包（内部的函数可以访问外部函数的变量，把函数返回出去。闭包可以保护内部的变量，闭包也可以造成内存泄露，==null.）
3. 原型链，
    1. 构造函数里的属性的优先级比原型链的要高
    2. 面向对象编程的时候， JS没有类的概念，可以用函数替代
    3. constructor 实际就是对应的那个函数
    4. prototype 按引用传递的 Object.create原型链的副本
4. 数值 字符串 布尔类型 按值传递
5. 改变this的方法 apply call bind
6. 函数提升 变量提升 函数提升的级别要比变量高
7. jq内部有很多经典的写法 模块化编程的概念 闭包

```js
function test(m) {
  m.v = 20;
}
var m = {age: 30};

test(m);
alert(m.v);
```

```js
// for 循环是同步
// 里面点击事件是异步
// 异步要等同步事件中的全部执行完成以后才执行。
var list_li = document.getElementsByTagName("li");
for (var i = 0; i < list_li.length; i++) {
  list_li[i].onclick = function() {
    console.log(i)
  }
}

// 解决
// 1. 闭包能够保护相应的变量
// 2. var -> let
```

```js
var module = (function() {
  var N = 5;
  function print() {
    console.log(N);
  }
  function add(x) {
    var q = x + N;
    console.log(q)
  }
  return {
    des: "这里是一个模块",
    add: add
  }
})();

module.add(3);

// 立即执行函数能够实现我们的模块化，立即执行函数就是一个模块话的东西。
```

```js
// 静态的模块
```
