### jQuery源码中形参undefined的意义所在

```js
(function(window,undefined){
})(window)
```

上面undefined是何意呢？
我们都知道undefined是一个全局原始值，代表为未定义的意思。

```js
// 在全局环境下
var undefined = 1;
console.log(undefined); // undefined
```

由此可见当在全局环境下定义一个变量为undefined的时候，无论把它变成什么值，都是undefined，我们都知道，js中查找变量是按照词法作用域查找的，如果jQuery中没有形参undefined。

```js
(function(window){
    undefined = 1;
    console.log(undefined); // undefined
})(window)
console.log(undefined) // undefined
```

这样的话 jQuery代码里面的undefined并没有改变，还是undefined，这就造成了污染全局变量的效果。如果jQuery的形参加上undefined

```js
(function(window,undefined){
    undefined = 1;
    console.log(undefined); // 1
})(window)
console.log(undefined); // undefined
```

### jQuery的$()方法是可以new的
使用过jQuery的都知道，当使用jQuery获取到某个元素的时候。是用jQuery的强大选择器直接选取的，但是又有谁知道jQuery的$()方法是可以new 的呢？

```js
var a = $("#box");
var b = new $("#box");
console.log(a);
console.log(b);
// a和b内容一模一样
```

上面一种是new的，一种不是new的，其实这两种作用都是一样的，选取的内容也是一样的。这是为什么呢？ 我们来深入jQuery的源码揭开它神秘的面纱。

```js
(function(window,undefined){
    var jQuery = function(seletor,context){
        return new jQuery.fn.init(selector,context);
    }
    jQuery.fn = jQuery.prototype = {
        init: function(seletor,context){
        }
    }
    jQuery.fn.init.prototype = jQuery.fn;
    window.$ = window.jQuery = jQuery;
})(window)
```

上面是jQuery核心的代码，在jQuery方法中可以看到，jQuery方法返回的一个new
jQuery.fn.init(selector,context); 这个到底是什么东西呢？先往下看

```js
jQuery.fn = jQuery.prototype = {
    init: function(seletor,context){
    }
}
```

这段代码我们发现jQuery.fn等于了jQuery.prototype，我们都知道对象是引用值，所以它们两个现在是同一个东西，jQuery.fn 也就是jQuery.prototype。同样我们看到jQuery.prototype中有个方法是init，而jQuery.prototype又等于jQuery.fn。这是我们就发现了在jQuery方法中返回的那个new jQuery.fn.init(selector,context)就是返回的 new jQuery.prototype.init(selector,context)。再往下看。
jQuery.fn.init.prototype = jQuery.fn;

这句话又是什么意思。我们来捋一下，由上面可得

> jQuery.fn = jQuery.prototype

得出

> jQuery.fn.init.prototype = jQuery.fn = jQuery.prototype;

当我们把中间的去掉之后

> jQuery.fn.init.prototype = jQuery.prototype;

哇，原来饶了这么一大圈只是为了让上面两个成等式。再看

> jQuery.fn.init = jQuery;

这时候我们就明白了为什么new和不new是一样的了，原来不new返回的jQuery.fn.init和new了之后返回的jQuery是一样的。

但是上面这样写是为了什么呢？

其实就是为了把`jQuery.fn.init.prototype`上面的方法全部都挂载上`jQuery.prototype`上。

### jQuery中的链式调用
jQuery中有一个很厉害的操作就是可以链式调用。

```js
$("#box").css("background",'red').attr('value',"hello").val();
```

以一个对象为例来解释jQuery中的链式调用

```js
var obj = {
    init:function(){
        console.log('init');
    },
    add : function() {
        console.log('add');
    },
    delete : function(){
        console.log('delete');
    }
}
obj.init() // 返回undefined
```

看到上面代码我们就想到如果这个init方法返回obj的话是不是就可以继续调用了，而我们又知道方法中存在this，this指向运行时决定的上下文。`obj.init()`此时init方法中的this就是指向obj。

```js
var obj = {
    init:function(){
        console.log('init');
        return this;
    },
    add : function() {
        console.log('add');
        return this;
    },
    delete : function(){
        console.log('delete');
        return this;
    }
}
obj.init().add().delete();  //init   add   delete
```

### jQuery中为创建元素绑定事件
jQuery最常见的就是操作Dom元素，当为创建的元素绑定事件的时候，使用click方法是不行的，这时用jQuery中的on方法绑定就可以。

```js
$("body").append("<div>hello , world</div>");
$("div").click(functon(){});  //这样是不行的
$('body').on("click",'div',function(){}); // 这样就是可以的。
```

这个on方法通过事件冒泡机制实现了事件委托，在body上绑定事件通过判定事件源来决定事件是否执行。那么什么是事件冒泡呢？

#### javascript中的事件流
在JavaScript中的事件流主要是通过事件捕获和事件冒泡实现的。例如：

```html
<body>
    <div>
        <p>
        </p>
    </div>
</body>
<script>
    document.addEventListener("click",function(){
    },true)  //最后一个参数为true就开启了事件捕获
</script>
```

### jQuery方法的重载
当我们在使用jQuery的时候，我们会发现，在使用jQuery方法的时候，参数不同所执行的功能就不同。例如：

```js
$("#box").val() //此时代表获取当前元素的value值
$("#box").val("hello");  //此时将传入的参数写入value值。
```

这就是jQuery中方法的重载，以参数个数的不同来执行不同的功能。

> jQuery是以闭包存储变量的特性非常巧妙的实现的。

```js
var addMethod = function(object,name,fn){
    var old = object[name] // 保存老的方法
    object[name] = function(){
        if(fn.length === arguments.length){//判断使用方法时传入的实参和定义此方法的形参是否相等。
            return fn.apply(this,arguments);
        } else {
            return old.apply(this,arguments);   //不符合执行老的方法
        }
    }
}
var obj = {
    name : ['张三','李四','王二'],
}
var find0 = function(){
    //不传入参数查找所有人的名字。
    return this.name;
}
var find1 = function(name){
    //传入参数name找到此人名的索引
    var arr = this.name;
    arr.forEach(function(item , index, arr){
        if(item === name) {
            console.log(index);
        }
    })
}
var find2 = function(name,age) {
    console.log(name,age);
}
addMethod(obj,'find',find0);
/*
    此时： old === undefined, fn === find0
*/
addMethod(obj,'find',find1); //注册同样的方法名
/*
    此时： old === find0, fn === find1
*/
addMethod(obj,'find',find2);
/*
    此时： old === find1, fn === find2
*/
console.log(obj.find()); //['张三','李四','王二']
console.log(obj.find('张三'));   //0
console.log(obj.find('李四',20)); //李四  20
```

查找方法方式从最近更新的那里开始查找，然后逐级往最小的查找。最终找到，因为最后一个find0就是没有参数，所有undefined被割舍掉。

### 短路表达式和多重短路表达式

短路表达式就是通过与或非逻辑运算来判断执行语句。多重短路表达式顾名思义就是多次与或非结合。

```
|| 物理中代表或门，当表达式两边有一边为true的时候执行，
    例如： a||b
        1.当a为false，b为false的时候，不通过
        2.当a为true，b为false的时候 ， 通过
        3.当a为false，b为true的时候， 通过
        4.当a为true，b为true的时候，通过
&& 物理中代表与门，当表达式两边都为true的时候执行
    例如 a && b
        1.当a为false，b为false的时候，不通过。
        2.当a为true，b为false的时候，不通过。
        3.当a为false，b为true的时候，不通过。
        4.当a为true，b为true的时候，通过

! 物理中代表非门，表示取反向的意思，
    例如： !a
        1.当a为false的时候，!a为true
        2.当a为true 的时候，!a为false
```

```js
var foo = a || b // 当a存在的时候把a赋予foo，当a不存在的时候把b赋予foo，当两者都不存在的时候等于undefined
```

例如：判断一个函数是否执行。

```js
var drag;
function test(){

}
drag && test() // 当drag
```

#### 多重短路表达式

```js
var foo = a && b && c //就相当于a && (b && c)
```

### jQuery中简化代码的方法

jQuery通过对象取得原型链上的方法，保存到局部变量，大大减少了通过原型链查询方法的时间。

```js
var class2type = {};
var toString = class2type.toString();
```

### jQuery中的类型判断

用过typeof的都知道它是不准的，通过Object.prototype.toString.call()来判断是正确的，但是太麻烦，jQuery通过对象的key-value做了一次映射，通过映射来判断什么类型。

```js
var class2type = {};
"Boolean Number String Function Array Date RegExp Object Error Math JSON Set Map Arguments".split(" ").map(function(item,index){
    class2type['[object ' + item +']'] = item.toLowerCase();
})
function type (obj) {
    if(obj == null) { // 利用undefined == null的特性，去除这两者的干扰
        return null + "";
    }
    return  typeof obj === "object" || typeof obj === "function" ? class2type[Object.prototype.toString.call(obj)] || 'object' : typeof obj;
}
```

### jQuery中的观察者模式

jQuery中通过on绑定事件，通过trigger触发事件，通过off接触事件实现观察者模式,同样可以自定义事件。

```js
$("#box1").on("someEvent",function(){ // 注册事件
    console.log('someEvent');
})
$("#box2").on("click",function(){
    $("#box1").trigger("someEvent");  发布时间
})
$("#box1").off('someEvent'); // 移除事件
```
