## PHP面向对象基础

### 面向对象的产生

面向对象的参数主要还是计算机史上软件危机的出现。人们为了解决软件危机从而诞生了软件工程学。

### 软件危机的产生

> 软件危机是指落后的软件生产方式无法满足迅速增长的计算机软件需求，从而导致软件开发与维护过程中出现一系列严重问题的现象。泛指在计算机软件的开发和维护过程中所遇到的一系列严重问题。

### 软件工程学

> 软件工程学士一门研究用工程化方法构建和维护有效的、实用的和高质量的软件的学科，它涉及到程序设计语言、数据库、软件开发工具、系统换平台、标准、设计模式等方法。分结构化方法（安软件周期分为三个阶段：分析、设计、编程）和面向对象。

### 面向对象

> OOP(Object-Oriented Programming，面向对象编程)技术为编程人员敞开了一扇大门，是其编程的代码更简洁、更易于维护，并且具有更高的课重用性。

> OOP达到了软件工程的三个目标：重用性、灵活性、扩展性。

> OOP面向对象编程的特点：封装、继承、多态。

#### 类与对象之间的关系

> 对象是类的实例，类是对对象的抽象。

#### 面向对象程序上的主要特征

1. 对象的行为
2. 对象的状态
3. 对象的标识

#### 如何抽象一个类

一个类有三部分是非常重要的：类的声明、成员属性、成员方法。

##### 类的声明

```php
// 简单格式
     [修饰符] class  类名{   //使用class关键字加空格后加上类名            修饰符表示这个类是私有的，还是公有的
            [成员属性]      //也叫成员变量
            [成员方法]      //也叫成员函数
    }
// 完整格式
     [修饰符] class 类名 [extends 父类] [implements 接口1[,接口2...]]{
            [成员属性]  //也叫成员变量
            [成员方法] // 也叫成员函数
     }
```

##### 成员属性

```php
// 格式：修饰符 $变量名[= 默认值]；例如：public $name = "张三";
// 注意：成员属性不可以是带运算符的表达式、变量、方法或函数调用。
    // 错误的形式
        public $var3 = 1 + 2; //错误
        public $var4 = self::myStaticMethod();      //错误
        public $var5 = $myVar;      //错误
    // 正确的形式
        public $var6 =100;      // 普通数值（4个标量：整数，浮点数，布尔，字符串）
        public $var6 = myConstant;  //常量
        public $var7 = self::classConstant // 静态属性，自己的属性
        public $var8 = array(true,false);   // 数组
```

##### 成员方法

```php
// 成员方法格式：
[修饰符] function 方法名(参数...) {
    [方法体]
    [return 返回值]
}
// 例如
    public function say(){      // 人可以说话的方法
        echo "hello world" ;        // 方法体
    }
```

#### 通过类实例化对象

##### 实例化对象

当定义好类后，我们使用new关键字来生成一个对象

```php
$对象名称 = new 类名称();
$对象名称 = new 类名称([参数列表]);
```

##### 对象中成员的访问

```php
// 语法：
    $引用名 = new 类名(构造参数)         实例对象
    $引用名->成员属性 = 赋值          对象属性赋值
    echo 引用名-> 成员属性        输出对象的属性
    $引用名->成员方法(参数)    调用对象的方法
```

##### 特殊对象引用$this

可以使用特殊对象引用$this,$this指向当前对象.

```php
// 例如
    public function play(){
        echo "正在玩手机";
    }
    public function info(){
        $this -> play();
        return "手机的宽度:{$this -> width}, 手机的高度：{$this -> height}";
    }
```

##### 实例

```php
class Person{
    public $age;
    public function say($word){
        echo "person say{$word}";
    }
    public function info(){
        $this -> say("Hi");
        return $this -> age;
    }
}
$xiaohong = new Person();
$xiaohong -> age = 22;
$age = $xiaohong ->info();
echo "</br>";
echo $age;
```

## 常见的关键字及方法

### final关键字

在PHP5中新增加了final关键字,它只能用来修饰类和方法.不能使用final这个关键字来修饰成员属性.

#### final的特性

1. 使用final关键字标识的类不能被继承
2. 使用final关键字标识的方法不能被子类覆盖(重写),是最终版本

#### 目的

1. 安全
2. 没必要被继承或重写

### static关键字

static关键字表示静态的意思,用于修饰类的成员属性和成员方法(即静态属性和静态方法).

类中的静态属性和方法不用实例化(new)就可以直接使用类名访问

> 格式： 类::静态属性 类::静态方法

在类的方法中。不能使用this来引用静态变量或静态方法，而需要用self来引用

> 格式：self::静态属性 self::静态方法

#### 特点

1. 静态方法中不可以使用非静态的内容。就是不让使用$this
2. 静态属性是共享的。也就是new很对对象也是共用一个属性。

### const关键字

const是在一个类中定义常量的关键字，我们都知道在PHP中定义常量使用的是“define()”这个函数，但是在类里面定义常量使用的是“const”这个关键字。

```php
const CONSTANT = "constant value"; // 定义
echo self ::CONSTANT // 类内部访问
echo className::CONSTANT // 类外部访问
```

### instanceof关键字

instanceof关键字用于检测当前对象实例是否属于某一个类或者这个类的子类。

### 自动加载类

PHP5当new实例化一个不存在的类时，则自动调用此函数__autoload(),并将类名作为参数传入此函数。可以使用这个实现类的自动加载。

### class_exists---检查类是否已定义

> 格式：bool class_exists(String $class_name[,bool $autoload])

如果由class_name所指的类已经定义，此函数返回TRUE，否则返回FALSE。

默认将会尝试调用__autoload，如果不想让class_exists()调用__autoload,可以将autoload参数设为FALSE。

### get_class ---返回对象的类名

> 格式：string get——返回对象的类名

返回对象实例obj所属类的名字。如果obj不是一个对象则返回FALSE

### get_object_vars--返回由对象属性组成的关联数组

> 格式：array get_object_vars(Object $obj)

返回由obj指定的对象中定义的属性组成的关联数组。

### get_parent_class -- 返回对象或类的父类名

> 格式：string get_parent_calss([mixed $obj])

如果obj是对象，则返回对象实例obj所属类的父类名

### is_a ---如果对象属于该类或该类是此对象的父类则返回TRUE

> 格式：bool is_a(object $object , string $class_name)

我们可以使用运算符：instanceof代替上面的is_a操作。

### __toString

当我们用echo输出一个引用对象的时候，会报错，如果在类中提供__toString方法，就不会产生错误，而自动调用__toString方法，输出__toString方法中返回的字符串，所以__toString方法一定要有个返回值（return语句）

```php
class Test{
    public function __toString(){
        return "hello";
    }
}
$test = new Test();
echo $test; // hello
```

### clone

根据一个对象完全可克隆出一个一模一样的对象，是非常有必要的，而且克隆以后，两个对象互不干扰。

```php
class Test{
    public $name = "liusixin";
}
$t1 = new Test();
$t2 = clone $t1;
$t2 -> name = "lsx";
echo $t1 -> name;
```

### 对象序列化

有时候需要吧一个对象在网络上传输，为了方便传输，可以把整个对象转化为二进制串，等到达另一端时，再还原为原来的对象，这个过程就叫做序列化（串行化）

#### 对象序列化的两种情况

1. 第一种情况就是把一个对象在网络中传输的时候要将对象串行化。
2. 第二种情况就是把对象写入文件或是数据库的时候用到串行化。

#### 两个过程

1. 串行化：就是把对象转化为二进制字符串，通过serialize()函数实现
2. 反串行化：就是把对象转化的二进制字符串再转为对象，通过unserialize()函数实现

```php
class Test{
    public $name = "liusixin";
    public function say(){
        return "hello";
    }
};
$test = new Test();
$testString = serialize($test);
echo $testString;
$untest = unserialize($testString);
echo $untest -> name; // liusixin
```

### __sleep()和————wakeup

当对象进行序列化的时候，调用类中的`__sleep()`方法，这个方法返回一个数组，数组中包含对象的属性，串行化只串行属猪中包含的属性，未被包含的则在串行化的时候被忽略，如果没有`__sleep()`方法，PHP将保存所有属性。

当对象进行反序列化的时候，调用类中的`__wakeup()`方法。

```php
class Test{
    public $name = "liusixin";
    public $age = 18;
    public $sex = "nan";
    public function __sleep(){
        $arr  = array('name','age');
        return $arr;
    }
    public function __wakeup(){
        $this -> age = 20;
    }
}
$test = new Test();
$testString = serialize($test);
echo $testString;
$test1 = unserialize($testString);
echo $test1 -> age; // 20
```

### __autoload()

在软件开发过程中，不可能把所有的类都写在一个php文件中，当一个文件调用另一个文件的时候就需要用include或者require,有时候一一导入所需要的类的php文件是一个很头疼的问题，所以__aotuload可以完成一个自动加载类，实现按需加载的形式。

在PHP5中，可以定义一`__autoload`函数，它会在视图使用尚未被定义的类时自动调用，通过调用次函数，脚本引擎在PHP出错失败前有个最后一次机会加载所有的类，`__autoload()`函数接受一个参数，就是想要加载的类名，通过拼接前缀或者后缀实现按需加载类.

```php
function __autoload($className) {
    require_once $className.".php";
}
try {
    $obj = new php1();
} catch (Exception $e){
    echo $e ->getMessage();
}
```

## 构造方法和析构方法

### 构造方法

当类被实例化的时候，构造方法调用

##### 构造方法的用途

用于实例化类的时候初始化相关数据。

```php
// 构造方法语法格式
    [修饰符] function__construct([参数]){
        程序体
    }
```

### 析构方法

当类不用再使用这个类里面的属性和方法的时候，析构方法执行。

##### 析构方法的用途

可以禁用资源的释放：数据库关闭、读取文件关闭等等，对象被销毁的时候执行，也就是说这个对象没有代码再运行了。

```php
// 析构方法方法格式
    [修饰符] function__destruct([参数]){
        程序体
    }
```

### 构造方法与析构方法的实例

```php
class Test {
    // 构造方法
    public function __construct($name , $age){
        echo 'hello'.$name;
        echo"</br>";
        $this -> age =$age;
        $this -> name = $name;
    }
    public function data(){
        return $this -> age;
    }
    // 析构方法。
    public function __destruct(){
        echo "bye bye {$this -> name}";
    }
}
$test = new Test('first',20);
$test = new Test('first',20);
// 对象的执行顺序----先进先出
// echo $test -> data();   //输出20
```

## 封装、继承、多态

### PHP面向对象之封装性

#### 设置私有成员与私有成员的访问

##### 封装的修饰符

封装性是面向对象编程中的三大特性之一，封装就是把对象中的成员属性和成员方法加上访问修饰符，使其尽可能隐藏对象的内部细节以达到对成员的访问控制（切记不是拒绝访问）。

```php
// PHP支持如下3种访问修饰符
    public     （公有的、默认）
    private      （私有的）
    protected     （受保护的）
```

##### 修饰符的访问权限


| 环境 | private | protected | public |
|:---:|:---:|:---:|:---:|
在同一类中|可以|可以|可以
在类的外部|不可以|不可以|可以

##### 设置私有成员

只要再声明成员或成员方法时，使用`private`关键字修饰就是实现了对成员的私有封装。

```php
class Person {
    private $name;
    private $age;
    private function say(){

    }
}
```

##### 访问私有成员

封装后的成员在对象的外部不能直接访问，只能在对象的内部方法中使用$this访问。

```php
class Person {
    private $name;
    public function say (){
        return $this -> name;
    }
}
```

#### 魔术方法

魔术方法只针对`protected`或者`private`的属性

##### __set()

```php
// set方法
public function __set($key,$value){
    // 魔术方法的set只针对 protected或者private的属性
    if($key == 'name' && $value == "laowang"){
        $this -> name = "xiaowang";
    }
}
$xw ->age = 18;
```

##### __get()

```php
// get方法
public function __get($key){
    if($key == 'age'){
        return 'girl not till you';
    }
}
echo $xw ->age;
```

##### __isset()

当成员属性可以在外部访问的时候isset判断为true、放成员属性无法在外部访问的时候isset判断为false .

```php
// isset方法
public function __isset($key){
    if($key == "age"){
        return true;
    }
}
var_dump(isset($xw -> age));        //false
// var_dump 打印输出
```

##### __unset()

把对象中的属性销毁

```php
// unset方法
public function __unset($key){
    if($key == "age") {
        unset($this -> age);
    }
}
unset($xw -> age);
```

### PHP面向对象之继承性

#### 类继承的应用

PHP只支持单继承,不允许多重继承.一个子类只能有一个父类,不允许一个子类直接继承多个类,但一个类可以被多个类继承，可以有多层继承,即一个雷可以继承某一个类的子类,如类B继承了类A,类C又继承了类B,那么类C也间接继承了类A。

```php
class A {

}
class B extends A{

}
```

#### 继承中修饰符的访问权限

| 环境 | private | protected | public |
|:---:|:---:|:---:|:---:|
在同一类中|可以|可以|可以
在类的外部|不可以|不可以|可以
在子类中|不可以|可以|可以

### PHP面向对象之多态

对象的多态：是指咋父类中定义的属性或行为被子类继承之后，可以具有不同的数据类型或表现出不同的行为。这使得同一个属性或方法在父类及其各个子类中具有不同的语义。

多态中最关键的实现形式就是函数的重载与重写。

#### 子类中重载父类的方法

1. 重载：父类中定义的方法，子类中有相同的方法名，但是参数个数不同
2. 重写：父类中定义的方法，子类中有相同的方法名，参数个数相同

在子类里面允许重写（覆盖）父类中的方法，使用parent访问父类中被覆盖的属性和方法

```php
parent::construct();
parent::fun();
```

#### 例子

```php
class Person {
    public $name;
    private $age;
    protected $money;
    public function __construct($name,$age , $money){
        $this -> name = $name;
        $this -> age = $age;
        $this -> money = $money;
    }
    public function cardInfo(){
        echo "name-> ".$this->name . "  age->".  $this->age . "   money". $this->money;
    }
}
class YellowPerson extends Person{
    function __construct($name,$age,$money){
        parent::__construct($name,$age,$money);
    }

    public function cardInfo($pp){
        //如果不加parent::cardInfo，就表示重写了父类的方法，再重新调用一下父类的方法。
        parent::cardInfo();
        echo $pp;
    }
    public function getMoney(){
        echo $this -> money;
    }
}
$s = new YellowPerson("xiaowang",22,100);
$s -> cardInfo(11);
echo $s -> name; // 公有属性可以继承
// echo $s -> age; // 私有属性无法继承
echo $s -> getMoney(); // 外部不能访问，但是可以继承过来
```

## 抽象类与接口

### 抽象方法和抽象类

当类中有一个方法,但没有方法体,也就是没有花括号,直接分号结束,像这种方法我们叫抽象方法,必须使用关键字abstract定义

```php
// 抽象类
abstrace class{
    // 抽象方法，抽象方法不用写方法体。
    public abstract function fun();
}
```

包含这种方法的类必须是抽象类也要使用关键字abstract加以声明。（即使用关键字abstract修饰的类为抽象类）。

#### 抽象类的特点

1. 不能实例化，也就是不能new成对象
2. 若想使用抽象类，就必须定义一个类去继承这个抽象类，并定义覆盖父类的抽象方法（实现抽象方法）。
3. 含有抽象方法的类必须是抽象类
4. 抽象类不一定非得含有抽象方法。
5. 抽象类可以存在普通的方法
6. 抽象类不能被实例化,必须由子类继承，并且把抽象类的抽象方法都实现。

#### 实例

```php
 abstract  class Person{
    //抽象方法不用写方法体。
    public abstract function eat();
}
class Man extends Person{
    public function eat(){
        echo "Man eat";
    }
}
$man = new Man();
$man ->eat();   //Man eat
```

### 接口技术

PHP与大多数面向对象编程语言一样，不支持多重继承，也就是说每一个类只能继承一个父类。为了解决这个问题，PHP引入了接口，接口的思想是指定了一个实现了该接口的类必须实现的一系列函数。

```php
定义格式:interface 接口名称{
    // 常量成员（使用const关键字定义）
    // 抽象方法（不需要使用abstract关键字）
}
使用格式:  class 类名 implements 接口1，接口2,{}
```

#### 接口的特点

1. 接口声明的关键字interface
2. 接口可以声明常量，也可以抽象方法
3. 接口中的方法都是抽象方法，不需要用abstract去定义。
4. 接口不能被实例化，需要一个类来实例化。
5. 一个类不能继承多个类，但是一个类可以实现多个接口

### 抽象类与接口的区别

#### 官方版

当关注一个事物的本质的时候，用抽象类；当你关注一个操作的时候，用接口。

#### 小官方版

接口是对动作的抽象，表示这个对象能做什么，对类的局部行为进行抽象。

抽象类是对根源的抽象，表示这个类什么，对类的整体进行抽象，对一类事务的抽象描述。

#### 人话版

比如：男人、女人，这两个类（如果是类的话.....），他们的抽象类是人（名词）。说明，他们都是人，人可以吃东西，狗可以吃东西，你可以把“吃东西”（动词）定义成一个接口，然后让这些类去实现它。所以，在高级语言上，一个类只能继承一个类（抽象类）（正如人不可能同时是生物和非生物），但是可以实现多个接口（吃饭接口，走路接口）。

#### 综上总结

1. 接口是抽象类的变体，接口中所有的方法都是抽象的。而抽象类是声明方法的存在而不去实现它的类。
2. 接口可以多继承，抽象类不行
3. 接口定义方法，不能实现，而抽象类可以实现部分方法。
4. 接口中基本数据类型为static 而抽象类不是
5. 接口中不能含有静态代码块以及静态方法，而抽象类可以函数静态方法和静态代码块。
6. 接口是纯粹的。

#### 实例

```php
 interface Person {
    const NAME = "liusixin";
    public function run();
    public function eat();
}
interface Study{
    public function study();
}
class Student implements Person,Study{
    // 类中self指自己
    const data = "3.13";
    public function run(){
        echo "run";
    }
    public function eat(){
        echo "eat";
    }
    public function study(){
        echo "study";
    }
    public static test(){
        echo self::data;
    }
}
$xw = new Student();
$xw -> eat();
$xw -> run();
$xw -> study();
// 使用接口中的静态的常量。
echo $xw::NAME;
```

## PHP异常处理

### 系统自带的异常处理

```php
class Exception{
    protected $message = "Unknown exception"; // 异常信息
    protected $code = 0; // 用户自定义异常代码
    protected $file; // 发生异常的文件名
    protected $line;
    function __construct($message = null ,$code = 0);
    final function getMessage(); // 返回异常信息
    final function getCode(); // 返回异常代码
    final function getFile(); // 返回发生异常的文件名
    final function getLine(); // 返回发生异常的代码行号
    final function getTrace(); // backtrace()数组
    final function getTraceAsString(); // 已格式化成字符串的getTrace()信息
    function __toString(); // 可输出的字符串。
}
```

#### 代码层面的使用

```php
// try分支在这里进行代码测试，如果有问题就要抛出一个异常，如果没有问题就继续执行
try {
    // catch分支就是捕捉异常对象
    // 参数：异常参数，使用的是类型约束，只能捕捉由Exception类实例化来的对象
} catch (Exception $e) {
    echo "错误文件为：";
    echo $e -> getFile(); // 得到发生异常的文件
    echo ",    发生错误的行为：";
    echo $e -> getLine(); // 得到发生错误的行为。
    echo ",    错误代码为：";
    echo $e ->getCode(); // 得到异常代码。
    echo ",   错误信息为：";
    echo $e -> getMessage(); // 得到异常信息。
}
```

### 自定义异常处理

自定义异常类时要继承系统的异常处理类

```php
class myException extends Exception {
    // 可以自己定义异常处理流程
    public function getAllInfo(){
        return "异常发生的文件为:{$this -> getFile()},异常发生的行为:{$this -> getLine()},异常的信息为:{$this -> getMessage()},异常的代码为:{$this->getCode()}";
    }
}
```

#### 代码层面的使用

```php
try {
    throw new myException("这是一个自定义的异常",123456);
} catch (myException $e) {
    // 捕捉时注意类型约束为自己定义的异常处理类名
    echo $e -> getAllInfo();
}
```

### 捕捉多个异常处理

在try代码之后，必须至少要给出一个catch代码块，也可以将多个catch代码块与一个try代码块关联使用。那么使用多个catch就可以捕获不同的类所产生的异常，注意顺序。

```php
try{
    // 捕捉多个异常处理要抛出多个异常对象，不能是由一个异常处理类实例化的对象
    if($_GET['num']==1){
        throw new myException('user');
    }else if{
        throw new Exception("sys");
    }
// 在捕捉时系统的异常处理分支要放在最后
// 注意类型约束
} catch (myException $e) {
    echo $e -> getAllInfo();
}catch(Exception $e){
    echo $e -> getMessage();
}
```
