###web安全
·代码层面
·架构层面
·运维层面

##安全问题
·用户身份被盗用
·用户密码泄露
·用户资料被盗取
·网站数据库泄露
·其它

##要点
·跨站脚本攻击xss（cross site scripting）
	·scripting能干啥？
		获取页面数据		偷取网站任意数据
		获取cookies		偷取用户资料
		劫持前端逻辑		偷取用户密码和登录态
		发送请求			欺骗用户
	·案例
		·站酷搜索（url搜索）
		·QQ空间（富文本）
		·XXX商城（后台）
	·XSS攻击分类
		·反射型（url参数直接注入）
		·存储型（存储到DB后读取时注入）
	·XSS攻击注入点
		·HTML节点内容（<div><script type="text/javascript">alert(1)</script></div>）
		·HTML属性(<img src="1" onerror="alert(1)" alt="">)
		·JavasSript代码(<script type="text/javascript">var data="#{data}";var data="hello";alert(1);"";</script>)
		·富文本
			·得保留HTML
			·HTML有XSS攻击风险
	·防御XSS
		·浏览器自带防御
			只针对反射型（html内容或html属性），并不是所有浏览器都支持
		·html转义（针对html标签和属性）
			var escapeHtml = function(str){
				if(!str) return '';
				str.replace(/&/g, '&amp;');//只能在最前面，会对下面做转义
				str.replace(/</g, '&lt;'); //<>html标签，引号html属性
				str.replace(/>/g, '&gt;');
				str = str.replace(/"/g, '&quto;');
				str = str.replace(/'/g, '&#39;');
				//str = str.replace(/ /g, '&#32;'); //空格可能有问题，对html属性最好加引号
				return str;
			}
		·针对js脚本
			<script type="text/javascript">
				var data="#{data}";
				var data="hello";alert(1);"";
			</script>
			var escapeForJs = function(str) {
				if(!str) return '';
				str = str.replace(/\\/g, '\\\\')
				str = str.replace(/"/g, '\\"')
			}
			用 JSON.stringify 转换
		·富文本（过滤，一般在输入的时候）
			·按黑名单过滤
				var xssFilter = function(html) {
					if(!html) return '';
					html = html.replace(/<\s*\/?script\s*>/g, '');
					html = html.replace(/javascript:[^'"]*/g, '');
					html = html.replace(/onerror\s*=\s*['"]?[^'"]*['"]?/g, '');
					return html;
				}
				// 标签太多，属性太多，不适合此办法
			·按白名单保留部分标签和属性
				cheerio库
				第三方库js-xss 	npm install xss
				var xssFilter = function(html) {
					if(!html) return '';
					var cheerio = require('cheerio');
					var $ = cheerio.load(html);

					// 白名单
					var whiteList = {
						'img': ['src'],
						'font': ['color', 'size'],
						'a': ['href']
					}
					$('*').each(function(index, elem){
						if(!whiteList[elem.name]){
							$(elem).remove();
							return;
						}

						for(var attr in elem.attribs) {
							if(whiteList[elem.name].indexOf(attr) === -1){
								$(elem).attr(attr, null);
							}
						}
					});
					console.log(html, $.html());
					return $.html();
				}
	·CSP（content security policy）最好解决方案
		·内容安全策略
		·用于指定哪些内容可执行
			·child-src connect-src default-src
			·font-src frame-src img-src
			·manifest-src media-src object-src
			·script-src style-src worker-src
			·<host-source> <scheme-source> 'self'
			·'unsafe-inline' 'unsafe-eval' 'none'
			·'nonce-<base64-value>' <hash-source>
			·'strict-dynamic'
	·xss防御需要注意的点（不同场景不同方法）
		·script标签的话转义 < > 符号
		·html属性中转义 引号
		·富文本中使用白名单方式
·跨站请求伪造攻击CSRF（cross site request forgy）
	·危害
		·利用用户登录态
		·用户不知情
		·完成业务请求
		·盗取用户资金
		·冒充用户发帖背锅
		·损坏网站名誉
	·原理
		第三方网站通过访问主网站后端带上cookie信息进行用户伪造
	·防御
		·禁止第三方网站带Cookies
			·same-site属性（同一站点）-但是只有chrome是完全可以
		·在前端页面加入验证信息（ccap模块生成验证码-npm install ccap --save）
			·验证码（攻击者网站不访问前端就没法拿到验证码通不过验证）
			·token
				var csrfToken = parseInt(Math.random()*9999999, 10);
		·验证referer
			·禁止来自第三方网站的请求 
			var referer = ctx.request.headers.referer;
			if(!/^https?:\/\/localhost/.test(referer))
			<!-- if(referer.indexOf('localhost') === -1){ -->
				throw new Error('非法请求')
			}
·前端cookie安全性
	·前端数据存储
	·后端通过http头设置
	·请求时通过http头传给后端
	·前端可读写
	·遵守同源策略（协议，域名，端口）
	·特性
		·域名
		·有效期
		·路径
		·http-only
		·secure（https）
		·samesite
	·作用
		·存储个性化设置
		·存储未登录时用户唯一标识
		·存储已登录用户的凭证
			·前端提交用户名和密码
			·后端验证用户名和密码
			·后端通过http头设置用户凭证
			·后续访问时后盾先验证用户凭证
		·存储其它业务数据
	·用户登录凭证
		·用户ID（隐患）
		·用户ID+签名
			// crypt.js
			var crypt = {};
			const KEY = '#ksdfidf312#!@%dj1'
			crypt.cryptUserId = function(userId){
				var crypto = require('crypto');
				var sign = crypto.createHmac = ('sha256', KEY);
				sign.update(userId + '');
				return sign.digest('hex');
			}
			module.exports = crypt;
			// 登录模块
			ctx.cookies.set('sign', crypt.cryptUserId(user.id), {
				httpOnly: false,
				sameSite: 'strict'
			});
			ctx.cookies.set('userId', crypt.cryptUserId(user.id), {
				httpOnly: false,
				sameSite: 'strict'
			});
		·SessionId
			session.js
			var session = {};
			var cache = {};
			session.set = function(userId, obj){
				var sessionId = Math.random();
				if(!cache[sessionId]){
					cache[sessionId] = {};
				}
				cache[sessionId].content = obj;
				return sessionId;
			}
			session.get = function(sessionId){
				return cache[sessionId] && cache[sessionId].content;
			}
			module.exports = session;
			site.js
			var sessionId = ctx.cookies.get('sessionId');
			var sessionObj = session.get(sessionId);
			if(!sessionObj || !sessionObj.userId){
				throw new Error('session不存在');
			}
			var userId = sessionObj.userId;
	·cookies和xss关系
		·xss可能偷取cookies
		·http-only的cookie不会被偷
	·cookies和csrf关系
		·csrf利用了用户cookies
		·攻击站点无法读写cookies
		·最好能阻止第三方使用cookies
	·cookies-安全策略
		·签名防篡改
		·私有变换（加密）- 信息隐藏
			var crypto = require('crypto');
			var KEY = '9851#%*!df1%!';
			var cipher = crypto.createCipher('des', KEY);
			var text = cipher.update('hello world', 'utf8', 'hex');//一边加密一边输出
			text += cipher.final('hex');
			console.log(text);

			var decipher = crypto.createDecipher('des', KEY);
			var originalText = decipher.update(text, 'hex', 'utf8');
			originalText += decipher.final('utf8');
			console.log(originalText);
		·http-only（防止xss）
		·secure（https）
		·same-site（主要csrf攻击，兼容性不好）
·点击劫持攻击
	·原理：通过iframe将目标网站放到页面上，将透明度设为0，直接点击的是iframe；
	·特征：
		·用户亲手操作
		·用户不知情
	·防御	
		·javascript禁止内嵌（有局限性）
			<script type="text/javascript">
				if(top.location != window.location){
					top.location = window.location;
				}
			</script>
			但是<iframe src="http://xxxx.cn:1211/post/15" sandbox="allow-forms"></iframe> //sandbox禁用功能
		·X-FRAME-OPTIONS 禁止内嵌（ie8以上）
			ctx.set('X-Frame-Options', 'DENY'); //SAME-ORIGIN同一网站，SAMEORIGIN嵌入页和被嵌入页在用一个域，ALLOW-FROM 网址 只允许指定网址进行内嵌
		·其他辅助手段
			·验证码
·传输过程安全问题
	·传输链路窃听篡改 
		traceroute www.baidu.com 查看详细访问节点
		代理软件 anyproxy
		·窃听
			·窃听用户密码
			·窃听传输敏感信息
			·非法获取个人信息
		·签名
			·插入广告
			·重定向网站
			·无法防御xss和csrf攻击
	·https
		浏览器 - 代理服务器 - 链路 - 服务器
				  TLS(SSL)加密（传输层加密）
		·中间人攻击
			浏览器 - 中间人 - 服务器
		·确定服务器身份
			·证书无法伪造
			·证书私钥不被泄露
			·域名管理权不泄露
			·CA坚守原则
		·证书
			https://www.sslforfree.com
		·SwitchHosts
		·https部署
			mkdir /opt/app/code/httpsweb
			cd /opt/app/code/httpsweb
			echo hello world > index.html
			vim /etc/nginx/conf.d/httpsweb.conf
				server{
					listen 80;
					listen 443 ssl https;
					server_name https.liusixin.cn;

					ssl_certificate /etc/letsencrypt/live/https.liusixin.cn/fullchain.pem;
					ssl_certificate_key /etc/letsencrypt/live/https.liusixin.cn/privkey.pem;

					location / {
						root /opt/app/code/httpsweb;
					}
				}
			curl https://get.acme.sh | sh
			cd /root/.acme.sh
			./acme.sh --issue -d https.liusixin.cn --webroot /opt/app/code/httpsweb
			修改conf证书地址
·用户密码安全问题
	·密码作用
	·密码存储
		·严禁明文存储 
			数据库：
				ALTER TABLE `USER` ADD COLUMN `salt` varchar(64) NULL DEFAULT "" AFTER `password`;
			user.js
				const results = await query(
					`select * from user where username = '${data.username}'`
				)
				if(results.length){
					let user = results[0];
					//如果用户没有salt
					//需要升级
					if(!user.salt){
						var salt = password.getSalt();
						var newPassword = password.encryptPassword(salt, user.password);
						await query(`update user set password = '${newPassword}', salt = '${salt}' where id = ${user.id}`);
					}
					var encryptPassword = password.encryptPassword(user.salt, data.password)
					if(encryptPassword !== user.password){
						throw new Error('密码不正确');
					}
				}
			password.js
				var password = {};
				var md5 = function(str){
					var crypto = require('crypto');
					var md5Hash = crypto.createHash('md5');
					md5Hash.update(str);
					return md5Hash.digest('hex'); 
				}
				password.getSalt = function() {
					return md5(Math.random()*999999+''+new Date().getTime());
				}
				password.encryptPassword = function(salt, password){
					return md5(salt + 'af$!#@$14#svc*&sf%' + password);
				}

				module.exports = password;
		·单向变换
			·彩虹表
				对抗彩虹表
				md5（明文）= 密文
				md5（md5（明文））= 密文
				md5（sha1（明文））= 密文
				md5（sha256（sha1（明文）））= 密文
				用复杂密码对抗彩虹表
				·帮助用户加强复杂度
					md5(sha1(md5(ID+ab83kd+原始密码+81kdso+盐+1lso;$2)))
				·变换次数越多越安全
					·加密成本几乎不变（生成密码时慢一些）
					·彩虹表失效（数量太大，无法建立通用性）
					·解密成本增大N倍
			·哈希算法
				·明文-密文 -- 对应
				·雪崩效应（明文有一点点不一样密文完全不一样）
				·密文 - 明文 无法反推
				·密文固定长度
				·常见哈希算法：md5 sha1 sha256
		·变换复杂度要求
		·密码复杂度要求
		·加盐
	·密码传输
		·https传输
		·频率限制
		·前端加密意义有限
	·密码替代方案
	·生物特征密码（指纹，虹膜，人脸）
		·指纹
		·声纹
		·虹膜
		·人脸
·SQL注入攻击
	·关系型数据库
		·存放结构化数据
		·可高效操作大量数据
		·方便处理数据之间的关联关系
		·常见：access/sqlite/mysql/mssql server
	·sql语言
		·语句
			select * from user where id=1;
			insert into user(username,password) values('test','123');
			update user set password='1234' where id=3;
			delete from user where id=3;
	·案例
		select * from user where username = 'TooBug' and password = '1' or '1'='1';否定前面语句
		select * from table where id="10" and mid(version(),1,1)=5 //截取版本号第一位是否等于5，等于5前面语句生效，否则报错。
		select id,1,2,3 from table;
		select * from table union select 1,2,3 from table2; // 联合查询
		select * from table where mid(username,1,1)="t" // 查询username第一位等于't';
	·注入攻击危害
		·猜解密码
		·获取数据
		·删库删表
		·拖库
	·sql注入防御
		·关闭错误输出
			catch(e){
				console.log('[/site/post] error:', e.message, e.stack);
				ctx.body = {
					status: e.code || -1,
					body: e.message
				}
			}
			改为
			ctx.body = {
				status: -1,
				body: '出错啦！'
			}
		·检查数据类型
			let id = ctx.params.id;
			id = parseInt(id, 10);
		·对数据进行转义
			let id = ctx.params.id;
			connection.escape(id);
		·使用参数化查询(抓包charles，wireshark)
			mysql2
			·原理：给一个意图和数据由系统自己去拼，而不是详细的sql语句
		·使用ORM(对象关系映射)- 增强安全性，开发效率
			·node中-sequelize 	npm install sequelize --save
			sequelize.js
			var Sequelize = require('sequelize');
			var sequelize = new Sequelize({
				host: 'localhost',
				database: 'safety',
				username: 'root',
				password: '123456',
				define: {
					freeezeTableName: true 
				}
			});
			module.exports = sequelize;
			post.js
			var sequelize = require('./sequelize');
			var Sequelize = require('sequelize');
			var Comment = sequelize.define('comment', {
				id: {
					type: Sequelize.INTERGER,
					primaryKey: true
				},
				userId: Sequelize.INTERGER,
				postId: Sequelize.INTERGER,
				content: Sequelize.TEXT
			}, {
				tableName: 'comment'
			})
			module.exports = Comment;
	·NoSQL注入和防御
		var mongoose = require('mongoose');
		login = async function(ctx){
			var username = ctx.request.body.username;
			var password = ctx.request.body.password;
			mongoose.findOne({
				username: username,
				password: password
			})
		}
		·检查数据类型
		·类型转换
		·写完整条件
·接入层上传问题
	·上传攻击
		·上传文件
		·再次访问上传的文件
		·上传的文件被当成程序解析
	·防御
		·限制上传后缀（比如php程序禁止上传php文件）
		·文件类型检查
			fileType
		·文件内容检查（开头特征）
			var fileBuffer = fs.readFileSync(file.path);
			fileBuffer[0] == 0x5b;
		·程序输出（node就是如此，但是因为要读到内存中，有读和写，性能受影响。折中办法：访问文件直接输出指令给到前端nginx服务器读取，性能损耗会小一些）
		·权限控制 - 可写可执行互斥（原则）
			上传文件可写不可执行
		·部署应用时用低权限用户，而不用root这种高权限用户
·信息泄露和社会工程学
	·信息泄露
		·泄露系统敏感信息
		·泄露用户敏感信息
		·泄露用户密码
	·信息泄露途径
		·错误信息失控（mysql）
		·sql注入
		·水平权限控制不当
		·xss/csrf
	·社会工程学
		昵称  ->   真名  -> 邮箱
		 ↓         ↓    ↘   ↓
	   手机号 -> 身份证号 -> 密码
	   ·你的身份由你掌握的资料确定
	   ·别人掌握了你的资料
	   ·别人伪装成你的身份
	   ·利用你的身份干坏事
	·案例
		·电信诈骗
		·伪装公检法
		·QQ视频借钱
		·微信伪装好友
	·OAuth思想
		QQ 			破网站

			  用户
		·一切行为由用户授权
		·授权行为不泄露敏感信息
		·授权会过期
	·利用OAuth思想防止资料泄露
			返回票据对应的数据>
		敏感资料 	  <->	业务
			  <带票据请求
				派发票据>
		敏感资料 	  <->	  用户
				 <登录
			 带票据请求>
		用户 	<-> 	业务

		·用户授权读取资料
		·无授权的资料不可读取
		·不允许批量获取数据
		·数据接口可风控审计
·其它安全问题
	·拒绝服务攻击DOS
		·特征
			·模拟正常用户
			·大量占用服务器资源
			·无法服务正常用户
		·类型
			·TCP半连接
			·HTTP连接
			·DNS（域名解析服务器）
		·大规模分布式拒绝服务攻击DDOS
			·流量可达几十到上百G
			·分布式（肉鸡、代理）
			·极难防御
		·DOS攻击案例
			·2009年5月19日 全国多省市断网
			· 游戏私服互相DDOS
			  换目标，攻击DNS服务器
			  DNS服务器机器下线
			  数十万网站DNS解析瘫痪
			  暴风影音后台疯狂请求解析
		·DOS攻击防御
			·防火墙
			·交换机、路由器
			·流量清洗
			·高防IP
		·DOS攻击预防
			·避免重逻辑业务
			·快速失败快速返回
			·防雪崩机制
			·有损服务
			·CDN
	·重放攻击
		·原理
			·请求被窃听或记录
			·再次发起相同的请求
			·产生以外的结果
		·后果
			·用户被多次消费
			·用户登录态被盗取
			·多次抽奖
		·防御
			·加密（https）
			·时间戳
			·token(session 后端生成)
			·nonce
			·签名
###总结
	·cookies和session关系
		session是基于cookies的存储来实现的（常用）
	·cookies特性
		·容量小，跟随html发送，通过http头设置，通过js读取
	·cookies属性
		·域名，有域名不能跨域
		·路径，同一域名path不一样cookies读取情况不一样
		·有效期
		·http-only，没办法通过js读取，通过http发送后台给后台读取
		·same-site
	·删除一个cookis值？
		设置一个过期时间为过去时间。
	·https如何保证数据不被窃听的
		·通过数据加密方式，在客户端和服务端进行通信之前来协商一个加密协议，通过加密协议来传输所有数据。
	·https如何保证不被中间人攻击
		·证书机制
	·部署https的步骤
		·确保全栈都可以部署到https
		·找ca生成证书
		·
###环境搭建
	npm install nodemon -g --registry=http://registry.npm.taobao.org
	npm install cheerio --registry=http://registry.npm.taobao.org -S
	导入数据库
		UPDATE mysql.user SET authentication_string=PASSWORD('xl941201@') WHERE User='root';
		flush privileges;
		mysql safety -uroot< safety.sql




