# 脚手架

- 快速生成配置。
- 降低框架学习成本。
- 令业务开发人员关注业务逻辑本身。

vue-cli (Vue 框架的命令行工具)就是解决这些问题的一个非常典型的例子。创建Vue项目的同时，vue-cli 提供了模板选择、编译以及本地开发服务器等功能模块。对于使用 vue-cli 创建的 Vue项目，业务开发人员无须操心复杂的 webpack 配置，只需关注业务逻辑开发本身，这很大程度上降低了开发的时间成本。

## 局限于本地的执行环境

前面讲述了前端工程化的 3 个阶段： 

- 本地工具链
- 云管理平台
- 持续集成

三者最明显的外在差异在于，对各个功能模块执行环境的划分。执行环境分为4类: 

- 本地环境 - 开发人员的本机环境
- 集成平台环境 - 云管理平台或者持续集成平台环境
- 测试环境 - 集成测试阶段测试工程师对产品进行仿真模拟测试的特定沙箱环境
- 生产环境 - 产品交付给用户的真实环境

不论工程化方案处于何种阶段，测试环境和生产环境的划分是一致的，因为两者与开发无关。

- 本地工具链阶段 - 这一阶段的工程化没有集成平台环境，脚手架、开发、本地服务器、构建以及部署测试功能模块全部在本地环境下执行。

![](http://cdn-blog.liusixin.cn/WX20180810-160454@2x.png)

- 云管理平台以及持续集成阶段 - 新增了集成平台环境，将构建与部署测试功能模块部署到云平台。而脚手架、开发以及本地服务器仍然在本地环境中执行

![](http://cdn-blog.liusixin.cn/WX20180810-160732@2x.png)

由此可见，不论前端工程化是最简单的本地工具链，还是集大成者的持续集成阶段，脚手架的执行环境始终局限于本地。这给脚手架工具带来了一个必须解决的问题：操作系统兼容性。

> 有的团队没有本机环境，而是提供虚拟机供开发人员使用。虚拟机可以规避由操作系统差异引起的额外开发成本，工程化工具也不必考虑操作系统兼容性问题。

## 多样性的实现模式

前端项目类型、资源的多样性以及各团队对前端工程师定位的不同造成前端脚手架没有固定的实现模式。不论具体实现模式如何，优秀的脚手架工具遵循的原则是一致的。

- 从功能实现的角度考量，需要具备：
  - 与构建、开发、部署等功能模块联动，在创建项目时生成对应配置项。
  - 自动安装依赖模块。
- 从平台角度考量，需要具备：
  - 动态可配置。
  - 底层高度可扩展。
- 从易用性角度考量，需要具备：
  - 丰富但不烦琐的配置项。
  - 支持多种运行环境，比如命令行和 Node.js APl。
  - 兼容各类主流操作系统。

## 开源脚手架案例剖析

- Sails.js - Node.js 全栈 MVC 框架。
- PHP中间层 - 只包括 Controller 和 View 的 Web 服务中间层框架，类似目前被广泛讨论的大前端。 
- Yeoman - 开放的脚手架平台，不封装任何具体方案。

### 1. Sails.js 一 针对服务器踹的脚手架方案

Sails.js 是一个企业级 Node.js 全栈框架，服务器端采用 MVC 架构，使用 Grunt 搭建前端工作流。Sails.js 的脚手架模块可以创建以下几种文件类型。

- `sails generate new app` - 创建一个 Sails.js 项目。
- `sails generate new model` - 创建一个 Model 文件。
- `sails generate new controller` - 创建一个 Controller 文件。
- `sails generate new api` - 创建一个API模块，包含一个 Model 文件和一个 Controller 文件。
- `sails generate new adapter` - 创建一个 Adapter 文件。Adapter 是 Database Adapter (数据库适配器)的简称，Sails.js 中的 Adapter 消除了各类数据库操作 API 的差异性，统一为 JavaScript 方法。
- `sails generate new generator` - 创建一个 generator 源码文件。

前5种类型都是 Sails.js 脚手架封装的具体方案，最后的 generator 类型相当于一种开发者 API。开发者可以通过这个入口封装符合团队业务需求的脚手架方案。另外，通过安装指定的脚手架插件，可以给 Sails.js 脚手架增加额外的配置项。

> Sails.js 的脚手架主要针对服务器端，并没有封装专门针对前端相关资源的方案，如果用户需要使用 Sails.js 脚手架创建前端相关模块则必须自行实现。

![](http://cdn-blog.liusixin.cn/WX20180810-162316@2x.png)

### 2. Yeoman 一 可能是最好的脚手架方案

面向 webapp 的脚手架工具，Yeoman 不能直接创建项目文件，它提供了一套完整的脚手架开发者 API, 使用这些 API 可以定制符合自己业务需求的任意脚手架方案。换句话说，Yeoman 不封装任何方案，它是完全开放、高度可扩展的。

![](http://cdn-blog.liusixin.cn/WX20180810-162749@2x.png)

## 集成 Yeoman 封装脚手架方案

脚手架创建项目文件的第一步是收集用户的配置信息，用户配置信息收集完成之后，脚手架下一步的工作是将这些动态的配置信息转 化成静态的文件内容。此阶段的转化工作通常使用 `Boilerplate` (样板文件，可以理解为HTML模板引擎)执行，Yeoman默认使用`EJS`引擎。

动态内容转化完成之后，如果有必要，可以将文件后缀类型修改为目标文件后缀类型，比如 Yeoman 将 `*.ejs`文件后缀修改为 `*.js`、`*.css`、`*.html`等。

最后一步便是将生成的文件复制到目标文件夹。

![](http://cdn-blog.liusixin.cn/WX20180810-163343@2x.png)

### 封装脚手架方案

使用 Yeoman 的命令行工具 `yo` 结合 `generator-generator` 模板创建一个空的脚手架模块文件。当然，你也可以按照规范手动创建文件。一个完整的空脚手架文件目录如下。

```shell
- app # 目录是源码文件
  - _prompts # 用于提示用户配置的选项内容
    - _html.js
    - _js.js
    - _style.js
  - index.js # 执行入口文件
  - templates # 脚手架所封装方案的项目文件源码
    - assets
    - boi-conf.ejs
    - gitignore
    - index.app.ejs
    - js
    - package.ejs
    - style
- package.json
```

封装一个 Yeoman 脚手架方案本质上是创建一个 Yeoman Generator 实例，从 ES6 引入的 Lass 角度也可以理解为创建了一个 Generator 子类。