# 四. 建造者模式

> 将一个复杂对象的构建层和表示层相分离，同样的构建过程可以采用不同的表示。

## 场景

工厂模式主要是用来创建对象的实例（简单工厂模式，工厂方法模式）或者是类簇（抽象工厂模式），关心的是最终的产出是什么，所以工厂模式我们得到的是对象的实例或者对象的类簇。然而建造者模式在创建对象的过程中则更为复杂一些。虽然目的也是为了创建对象，但是更关心的创建对象的整个过程或者说是每一个细节。

比如创建一个人，我们创建的结果不仅仅是得到一个人的实例，还要关注创建人的时候，这个人是男是女，穿什么衣服带什么帽子等等。

```js
var Human = function(param) {
  this.skill = (param && param.skill) || '保密';
  this.hobby = (param && param.hobby) || '保密';
};
Human.prototype = {
  constructor: Human,
  getSill: function() {
    return this.skill;
  },
  getHobby: function() {
    return this.hobby;
  }
};
var Name = function(name) {
  var that = this;
  (function(name, that) {
    this.wholeName = name;
    if (name.indexOf(' ') > -1) {
      that.firstName = name.slice(0, name.indexOf(' '));
      that.secondName = name.slice(name.indexOf(' '));
    }
  })(name, that);
};
var Work = function(work) {
  var that = this;
  (function(work, that) {
    switch (work) {
      case 'code':
        that.work = '工程师';
        that.wordDesc = '代码使我快乐';
        break;
      case 'UE':
        that.work = '设计师';
        that.wordDesc = '设计更似艺术';
        break;
      default:
        that.work = work;
        that.wordDesc = '对不起，我们还不清楚你所选择职位的相关描述';
    }
  })(work, that);
};

//更换期望职位以及描述
Work.prototype.changeWork = function(work) {
  this.work = work;
};

Work.prototype.changeDesc = function(desc) {
  this.wordDesc = desc;
};

//创建一个应聘者
var Person = function(name, work) {
  var _person = new Human();
  _person.name = new Name(name);
  _person.work = new Work(work);
  return _person;
};

var person = new Person('Liusixin', 'code');
console.log(person.skill);
console.log(person.hobby);
console.info(person.work);
person.work.changeDesc('一撸代码就疯狂');
console.info(person.work);
```