# 二. 适配器模式

> 将一个类的接口转换为另外一个类的接口以满足用户的需求，使类之间的接口不兼容问题通过适配器得以解决。

我们来举一个例子，鸭子（Dock）有飞（fly）和嘎嘎叫（quack）的行为，而火鸡虽然也有飞（fly）的行为，但是其叫声是咯咯的（gobble）。如果你非要火鸡也要实现嘎嘎叫（quack）这个动作，那我们可以复用鸭子的 quack 方法，但是具体的叫还应该是咯咯的，此时，我们就可以创建一个火鸡的适配器，以便让火鸡也支持 quack 方法，其内部还是要调用 gobble。

首先要先定义鸭子和火鸡的抽象行为，也就是各自的方法函数：

```js
//鸭子
var Duck = function() {};
Duck.prototype.fly = function() {
  throw new Error('该方法必须被重写!');
};
Duck.prototype.quack = function() {
  throw new Error('该方法必须被重写!');
};

//火鸡
var Turkey = function() {};
Turkey.prototype.fly = function() {
  throw new Error(' 该方法必须被重写 !');
};
Turkey.prototype.gobble = function() {
  throw new Error(' 该方法必须被重写 !');
};

//鸭子
var MallardDuck = function() {
  Duck.apply(this);
};
MallardDuck.prototype = new Duck(); //原型是Duck
MallardDuck.prototype.fly = function() {
  console.log('可以飞翔很长的距离!');
};
MallardDuck.prototype.quack = function() {
  console.log('嘎嘎！嘎嘎！');
};

//火鸡
var WildTurkey = function() {
  Turkey.apply(this);
};
WildTurkey.prototype = new Turkey(); //原型是Turkey
WildTurkey.prototype.fly = function() {
  console.log('飞翔的距离貌似有点短!');
};
WildTurkey.prototype.gobble = function() {
  console.log('咯咯！咯咯！');
};
```

为了让火鸡也支持 quack 方法，我们创建了一个新的火鸡适配器`TurkeyAdapter`：

```js
var TurkeyAdapter = function(oTurkey) {
  Duck.apply(this);
  this.oTurkey = oTurkey;
};
TurkeyAdapter.prototype = new Duck();
TurkeyAdapter.prototype.quack = function() {
  this.oTurkey.gobble();
};
TurkeyAdapter.prototype.fly = function() {
  var nFly = 0;
  var nLenFly = 5;
  for (; nFly < nLenFly; ) {
    this.oTurkey.fly();
    nFly = nFly + 1;
  }
};
```

该构造函数接受一个火鸡的实例对象，然后使用 Duck 进行 apply，其适配器原型是 Duck，然后要重新修改其原型的 quack 方法，以便内部调用 oTurkey.gobble()方法。其 fly 方法也做了一些改变，让火鸡连续飞 5 次（内部也是调用自身的 oTurkey.fly()方法）。

```js
var oMallardDuck = new MallardDuck();
var oWildTurkey = new WildTurkey();
var oTurkeyAdapter = new TurkeyAdapter(oWildTurkey);

//原有的鸭子行为
oMallardDuck.fly();
oMallardDuck.quack();

//原有的火鸡行为
oWildTurkey.fly();
oWildTurkey.gobble();

//适配器火鸡的行为（火鸡调用鸭子的方法名称）
oTurkeyAdapter.fly();
oTurkeyAdapter.quack();
```